"""Utilities for core."""


def is_(self, state):
    return self.state == state


def set_(self, state):
    attr = self._meta['state_attribute_name']
    state = self._meta['reversed_states_map'][state]
    setattr(self, attr, state)


def state_getter(self):
    attr = self._meta['state_attribute_name']
    try:
        return getattr(self, attr).value
    except AttributeError:
        return None


def state_setter(self, value):
    attr = self._meta['state_attribute_name']
    return setattr(self, attr, value)


def state_deleter(self):
    pass


def generate_checker(value):

    @property
    def checker(self):
        return self.is_(value)

    return checker


def generate_setter(value):

    def setter(self):
        self.set_(value)

    return setter

state_property = property(state_getter, state_setter, state_deleter)



"""State machine core."""

from enum import Enum, unique
from functools import partial

import six

from . import utils

def _get_config(meta, attr):
    try:
        return getattr(meta, attr)
    except AttributeError:
        return getattr(DefaultMeta, attr)


class DefaultMeta(object):

    states_enum_name = 'States'


class StateMachineMetaclass(type):

    def __new__(cls, name, bases, attrs):
        new_class =  super(cls, cls).__new__(cls, name, bases, attrs)

        parents = [b for b in bases if isinstance(b, cls)]
        if not parents:
            return new_class

        meta = getattr(new_class, 'Meta', DefaultMeta)
        get_config = partial(_get_config, meta)
        new_meta = {}

        states_enum_name = get_config('states_enum_name')
        states_enum = getattr(new_class, states_enum_name)
        states_enum = unique(states_enum)

        new_meta['states_enum'] = states_enum
        new_meta['reversed_states_map'] = {s.value: s for s in states_enum}

        state_name = 'state'
        new_state_name = '_' + state_name
        new_meta['state_attribute_name'] = new_state_name
        state_value = None

        try:
            state_value = getattr(new_class, state_name)
            if isinstance(state_value, Enum):
                if not state_value in states_enum:
                    raise ValueError(
                        'Initial state does not belong to states enum!')
            else:
                raise ValueError(
                    'Initial value cannot be scalar, use Enum instance!')

        except AttributeError:
            pass

        setattr(new_class, new_state_name, state_value)
        setattr(new_class, state_name, utils.state_property)

        new_methods = {}
        for s in states_enum:
            checker_name = 'is_{}'.format(s.value)
            new_methods[checker_name] = utils.generate_checker(s.value)

            setter_name = 'set_{}'.format(s.value)
            new_methods[setter_name] = utils.generate_setter(s.value)

        for name, method in new_methods.items():
            if hasattr(new_class, name):
                raise ValueError(
                    "Name collision in generated class - '{}'.".format(name))

            setattr(new_class, name, method)

        setattr(new_class, '_meta', new_meta)
        setattr(new_class, 'is_', utils.is_)
        setattr(new_class, 'set_', utils.set_)

        return new_class


class StateMachine(six.with_metaclass(StateMachineMetaclass)):

    pass



import unittest
from enum import Enum

from super_state_machine import machine


class TestSuper_state_machine(unittest.TestCase):

    def test_state_machine_state_is_none(self):

        class Machine(machine.StateMachine):

            class States(Enum):

                ONE = 'one'
                TWO = 'two'
                THREE = 'three'

        sm = Machine()
        self.assertIsNone(sm.state)

    def test_state_machine_is_always_scalar(self):

        class Machine(machine.StateMachine):

            class States(Enum):

                ONE = 'one'
                TWO = 'two'
                THREE = 'three'

            state = States.ONE

        sm = Machine()
        self.assertEqual(sm.state, 'one')

    def test_state_machine_doesnt_allow_scalars_on_init(self):
        try:

            class Machine(machine.StateMachine):

                class States(Enum):

                    ONE = 'one'
                    TWO = 'two'
                    THREE = 'three'

                state = 'two'

        except ValueError:
            return

        raise RuntimeError('ValueError should be raised.')

    def test_state_machine_accepts_enums_only_from_proper_source(self):

        class OtherEnum(Enum):

            FOUR = 'four'

        try:

            class Machine(machine.StateMachine):

                class States(Enum):

                    ONE = 'one'
                    TWO = 'two'
                    THREE = 'three'

                state = OtherEnum.FOUR

        except ValueError:
            return

        raise RuntimeError('ValueError should be raised.')

    def test_state_machine_doesnt_allow_wrong_scalars(self):
        try:

            class Machine(machine.StateMachine):

                class States(Enum):

                    ONE = 'one'
                    TWO = 'two'
                    THREE = 'three'

                state = 'four'

        except ValueError:
            return

        raise RuntimeError('ValueError should be raised.')

    def test_state_machine_accepts_only_unique_enums(self):
        try:

            class Machine(machine.StateMachine):

                class States(Enum):

                    ONE = 'one'
                    TWO = 'one'
                    THREE = 'three'

        except ValueError:
            return

        raise RuntimeError('ValueError should be raised.')

    def test_state_machine_allows_to_change_and_check_state(self):

        class Machine(machine.StateMachine):

            class States(Enum):

                ONE = 'one'
                TWO = 'two'
                THREE = 'three'

        sm = Machine()
        self.assertIsNone(sm.state)
        self.assertIs(sm.is_('one'), False)
        sm.state = Machine.States.ONE
        self.assertIs(sm.is_('one'), True)
        self.assertEqual(sm.state, 'one')
        self.assertIs(sm.is_('two'), False)
        sm.set_('two')
        self.assertIs(sm.is_('two'), True)


    def test_state_machine_allows_to_change_and_check_state_by_methods(self):

        class Machine(machine.StateMachine):

            class States(Enum):

                ONE = 'one'
                TWO = 'two'
                THREE = 'three'

        sm = Machine()
        self.assertIsNone(sm.state)
        self.assertIs(sm.is_one, False)
        sm.state = Machine.States.ONE
        self.assertIs(sm.is_one, True)
        self.assertEqual(sm.state, 'one')
        self.assertIs(sm.is_two, False)
        sm.set_two()
        self.assertIs(sm.is_two, True)

    def test_name_collistion_for_checker_raises_exception(self):
        try:

            class Machine(machine.StateMachine):

                class States(Enum):

                    ONE = 'one'
                    TWO = 'two'
                    THREE = 'three'

                def is_one(self):
                    pass

        except ValueError:
            return

        raise RuntimeError('ValueError should be raised.')

    def test_name_collistion_for_setter_raises_exception(self):
        try:

            class Machine(machine.StateMachine):

                class States(Enum):

                    ONE = 'one'
                    TWO = 'two'
                    THREE = 'three'

                def set_one(self):
                    pass

        except ValueError:
            return

        raise RuntimeError('ValueError should be raised.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='super_state_machine',
    version='0.1.0',
    description='Super State Machine gives you utilities to build finite state machines.',
    long_description=readme + '\n\n' + history,
    author='Szczepan Cieślik',
    author_email='szczepan.cieslik@gmail.com',
    url='https://github.com/beregond/super_state_machine',
    packages=[
        'super_state_machine',
    ],
    package_dir={'super_state_machine':
                 'super_state_machine'},
    include_package_data=True,
    install_requires=[
        'enum34',
        'six',
    ],
    license="BSD",
    zip_safe=False,
    keywords='super_state_machine',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
)
